import json
import os
from typing import Dict
from typing import List
from uuid import UUID

from langchain.schema import Document

from opencopilot.logger import api_logger

logger = api_logger.get()

DEFAULT_CONVERSATION_LOGS_DIR = "conversation_logs"


class ConversationLogsRepositoryLocal:

    def __init__(self, conversation_logs_dir: str = DEFAULT_CONVERSATION_LOGS_DIR):
        os.makedirs(conversation_logs_dir, exist_ok=True)
        self.conversation_logs_dir = conversation_logs_dir

    def log_prompt_template(
            self,
            conversation_id: UUID,
            message: str,
            prompt_template: str,
            response_message_id: str,
            token_count: int = None,
    ) -> None:
        self._append_to_file(conversation_id, {
            "response_message_id": response_message_id,
            "message": message,
            "prompt_template": prompt_template,
            "token_count": token_count,
        })

    def log_prompt_text(
            self,
            conversation_id: UUID,
            message: str,
            prompt_text: str,
            response_message_id: str,
            token_count: int = None,
    ) -> None:
        self._append_to_file(conversation_id, {
            "response_message_id": response_message_id,
            "message": message,
            "prompt_text": prompt_text,
            "token_count": token_count,
        })

    def log_history(
            self,
            conversation_id: UUID,
            message: str,
            history: str,
            response_message_id: str,
            token_count: int = None,
    ) -> None:
        self._append_to_file(conversation_id, {
            "response_message_id": response_message_id,
            "message": message,
            "history": history,
            "token_count": token_count,
        })

    def log_context(
            self,
            conversation_id: UUID,
            message: str,
            contexts: List[Document],
            response_message_id: str,
            token_count: int = None,
    ) -> None:
        self._append_to_file(conversation_id, {
            "response_message_id": response_message_id,
            "message": message,
            "context": json.dumps([c.dict() for c in contexts]),
            "token_count": token_count,
        })

    def get_logs_by_message(
            self,
            conversation_id: UUID,
            response_message_id: str,
    ) -> List[Dict]:
        try:
            with open(self._get_file_path(conversation_id), "r") as file:
                lines = file.readlines()
            formatted_lines: List[Dict] = []
            for line in lines:
                if not isinstance(line, str):
                    continue
                formatted_line = json.loads(line)
                if formatted_line.get("response_message_id") == response_message_id:
                    formatted_lines.append(formatted_line)
            return formatted_lines
        except:
            return []

    def _get_file_path(self, conversation_id: UUID):
        return os.path.join(self.conversation_logs_dir, str(conversation_id)) + ".jsonl"

    def _append_to_file(self, conversation_id: UUID, log: Dict) -> None:
        try:
            with open(self._get_file_path(conversation_id), "a") as file:
                file.write(json.dumps(log) + "\n")
        except:
            logger.error(f"Error log_prompt_text, conversation_id: {str(conversation_id)}")



from opencopilot.domain.evaluation import evaluate_prediction_use_case
from opencopilot.service.debug.entities import EvaluationInput
from opencopilot.service.debug.entities import EvaluationResponse


async def execute(request: EvaluationInput) -> EvaluationResponse:
    result = await evaluate_prediction_use_case.execute(
        EvaluationInput(
            query=request.query,
            answer=request.answer,
            expected_answer=request.expected_answer,
        )
    )
    return EvaluationResponse(
        response="OK",
        evaluation=result.evaluation,
    )



import csv
import json
import os
import re
from typing import List

from opencopilot import settings
from langchain.document_loaders import CSVLoader
from langchain.document_loaders import PyPDFLoader
from langchain.document_loaders import TextLoader
from langchain.document_loaders import UnstructuredExcelLoader
from langchain.document_loaders import UnstructuredFileLoader
from langchain.schema import Document
from opencopilot.logger import api_logger

logger = api_logger.get()


def execute(
        data_dir: str,
        is_loading_deprecated=False,
        text_splitter=None
) -> List[Document]:
    files = []
    if not data_dir or not os.path.isdir(data_dir):
        return []
    ignore_files = ['.DS_Store']
    for (dir_path, dir_names, file_names) in os.walk(data_dir):
        for file_name in file_names:
            if file_name not in ignore_files:
                files.append(os.path.join(dir_path, file_name))

    files_count = len(files)
    if files_count > 0:
        logger.info(f"[Loading {files_count} file{'s'[:files_count ^ 1]} from {data_dir}.]")
    else:
        raise Exception(f"[Add files to {data_dir}.]")

    documents = []
    for file_path in files:
        new_documents = []
        if (file_size := _get_file_size(file_path)) > settings.get().MAX_DOCUMENT_SIZE_MB:
            logger.error(
                f"Document {file_path} too big ({file_size} > {settings.get().MAX_DOCUMENT_SIZE_MB}), skipping.")
            continue
        if file_path.endswith('.csv'):
            with open(file_path, newline='') as csvfile:
                dialect = csv.Sniffer().sniff(csvfile.read(1024))

            loader = CSVLoader(file_path=file_path, csv_args={
                'delimiter': dialect.delimiter,
            })
            new_documents = loader.load()
        elif file_path.endswith('.tsv'):
            loader = CSVLoader(file_path=file_path, csv_args={
                'delimiter': '\t',
            })
            new_documents = loader.load()
        elif file_path.endswith('.pdf'):
            loader = PyPDFLoader(file_path)
            new_documents = loader.load()
        elif file_path.endswith('.xls') or file_path.endswith('.xlsx'):
            loader = UnstructuredExcelLoader(file_path)
            new_documents = loader.load()
        elif (file_path.endswith(".json")
              and file_path.replace(data_dir, "").startswith("serialized_documents_")):
            with open(file_path, "r") as f:
                document_dicts = json.load(f)
            for document in document_dicts:
                metadata = document["metadata"]
                if metadata["source"] in settings.get().copilot_config.data.get("ignore", []):
                    continue
                if "deprecated" in metadata["source"]:
                    if not is_loading_deprecated:
                        # skip deprecated information
                        continue
                    document["metadata"]["deprecated"] = True
                content = _cleanup_document(document["page_content"])
                new_documents.append(Document(page_content=content, metadata=metadata))
        elif file_path.endswith(".json"):
            loader = TextLoader(file_path)
            new_documents = loader.load()
        else:
            try:
                loader = UnstructuredFileLoader(file_path)
                new_documents = loader.load()
            except Exception as e:
                logger.error(f"Error loading {file_path}", exc_info=True)
        if text_splitter:
            document_chunks = []
            for document in new_documents:
                for chunk in text_splitter.split_text(document.page_content):
                    document_chunks.append(Document(page_content=chunk, metadata=document.metadata))
            logger.info(
                f"Generated {len(document_chunks)} document chunks from {len(new_documents)} documents")
            documents.extend(document_chunks)
        else:
            logger.info(f"Generated {len(new_documents)} documents")
            documents.extend(new_documents)
    return documents


def _cleanup_document(text: str) -> str:
    # cleanup special unicode characters + trailing newline
    pattern = r"[\u00A0-\uFFFF]\n?"
    clean = re.sub(pattern, "", text)
    clean = re.sub("\n+", "\n", clean)
    return clean


def _get_file_size(file_path):
    size_in_bytes = os.path.getsize(file_path)
    size_in_mb = size_in_bytes / 1024 / 1024
    return size_in_mb



from typing import Any
from typing import List
from typing import Optional

import os
from urllib import parse
from bs4 import BeautifulSoup
from langchain.document_loaders import GitbookLoader
from langchain.schema import Document


def execute(urls: List[str]) -> List[Document]:
    documents = []
    for url in urls:
        loader = CustomGitbookLoader(url, load_all_paths=True)
        documents.extend(loader.load())
    return documents


class CustomGitbookLoader(GitbookLoader):

    def __init__(
        self,
        web_page: str,
        load_all_paths: bool = False,
        base_url: Optional[str] = None,
        content_selector: str = "main",
    ):
        super().__init__(web_page, load_all_paths, base_url, content_selector)
        self.parsed_url = parse.urlparse(web_page)

    def _get_document(
            self, soup: Any, custom_url: Optional[str] = None
    ) -> Optional[Document]:
        """Fetch content from page and return Document."""
        page_content_raw = soup.find(self.content_selector)
        i = 0
        last_child: Optional[Any] = None
        for child in page_content_raw.children:
            last_child = child
            i = i + 1
        if i > 2:
            print("Unexpected page content")
        else:
            """
            Pages have this format:
            <main>
                <div>title</div>
                <div>
                    <div>content here</div>
                    <div>navigation and other info</div>
                </div>
            </main>
            This removes the navigation part from the html
            """
            if last_child:
                i = 0
                for child in last_child.children:
                    if i == 1:
                        child.decompose()
                    i = i + 1
        if not page_content_raw:
            return None

        # Find all the 'a' tags in the HTML and convert them into Markdown links
        for link in page_content_raw.find_all("a"):
            href = link.get("href")
            if href is not None and link.text is not None:
                if href.startswith("/"):
                    absolute_path = os.path.join(self.parsed_url.netloc, href.lstrip("/"))
                    href = parse.urlunparse((self.parsed_url.scheme, absolute_path, '', '', '', ''))
                link_md = f"[{link.text}]({href})"
                link.replace_with(BeautifulSoup(link_md, "html.parser"))

        content = page_content_raw.get_text(separator="\n").strip()
        title_if_exists = page_content_raw.find("h1")
        title = title_if_exists.text if title_if_exists else ""
        metadata = {"source": custom_url or self.web_path, "title": title}
        return Document(page_content=content, metadata=metadata)



from starlette import status


class APIErrorResponse(Exception):
    """Base class for other exceptions"""

    def __init__(self):
        pass

    def to_status_code(self) -> status:
        raise NotImplementedError

    def to_code(self) -> str:
        raise NotImplementedError

    def to_message(self) -> str:
        raise NotImplementedError


InternalServerErrorCode = "internal_server_error"
InternalServerErrorMessage = "The request could not be completed due to an " \
                             "internal server error."


class InternalServerAPIError(APIErrorResponse):
    """Raised when an internal server error occurs"""

    def __init__(self):
        pass

    def to_status_code(self) -> status:
        return status.HTTP_500_INTERNAL_SERVER_ERROR

    def to_code(self) -> str:
        return InternalServerErrorCode

    def to_message(self) -> str:
        return InternalServerErrorMessage


class ValidationAPIError(APIErrorResponse):

    def __init__(self, param: str):
        self.param = param

    def to_status_code(self) -> status:
        return status.HTTP_422_UNPROCESSABLE_ENTITY

    def to_code(self) -> str:
        return "validation_error"

    def to_message(self) -> str:
        return f"Invalid parameter: {self.param}"


class NotFoundAPIError(APIErrorResponse):
    def __init__(self, resource: str):
        self.resource = resource

    def to_status_code(self) -> status:
        return status.HTTP_404_NOT_FOUND

    def to_code(self) -> str:
        return "not_found"

    def to_message(self) -> str:
        return f"{self.resource} not found."


class BadGatewayAPIError(APIErrorResponse):
    """Raised when an error occurs on the upstream server"""

    def __init__(self):
        pass

    def to_status_code(self) -> status:
        return status.HTTP_502_BAD_GATEWAY

    def to_code(self) -> str:
        return "bad_gateway"

    def to_message(self) -> str:
        return "The request could not be completed due to the server " \
               "receiving an invalid response from an inbound server it " \
               "accessed while attempting to fulfill the request. " \
               "Check if LLM is running."


class AuthorizationMissingAPIError(APIErrorResponse):

    def __init__(self):
        pass

    def to_status_code(self) -> status:
        return status.HTTP_401_UNAUTHORIZED

    def to_code(self) -> str:
        return "authorization_missing"

    def to_message(self) -> str:
        return f"Request is missing or has invalid 'Authorization' header."


class InvalidCredentialsAPIError(APIErrorResponse):

    def __init__(self):
        pass

    def to_status_code(self) -> status:
        return status.HTTP_401_UNAUTHORIZED

    def to_code(self) -> str:
        return "invalid_credentials"

    def to_message(self) -> str:
        return "Invalid credentials"



from typing import Optional

import jwt
from fastapi import Security
from fastapi.security import APIKeyHeader

from opencopilot import settings
from opencopilot.service import error_responses

API_KEY_NAME = "Authorization"
API_KEY_HEADER = APIKeyHeader(name=API_KEY_NAME, auto_error=False)


async def execute(api_key_header: str = Security(API_KEY_HEADER)) -> Optional[str]:
    if not settings.get().AUTH_TYPE:
        return None

    if not api_key_header:
        raise error_responses.AuthorizationMissingAPIError
    if api_key_header.startswith("Bearer "):
        return await _validate_jwt(api_key_header[7:])
    if api_key_header != settings.get().API_KEY:
        raise error_responses.AuthorizationMissingAPIError


async def _validate_jwt(payload: str) -> str:
    try:
        payload = jwt.decode(payload, settings.get().JWT_CLIENT_SECRET, algorithms=['HS256'])
        return payload.get("sub")
    except:
        raise error_responses.AuthorizationMissingAPIError



import logging
import os
from typing import Optional

from opencopilot import settings
from uuid import UUID
from pythonjsonlogger import jsonlogger

LOGGING_MESSAGE_FORMAT = "%(asctime)s %(name)-12s %(levelname)s %(message)s"

logger: Optional[any] = None


def get(agent_id: UUID = None):
    global logger
    if logger:
        return logger
    name = settings.get().APPLICATION_NAME
    file_handler = get_file_logger()
    console_handler = get_console_logger()
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    logger.addHandler(console_handler)
    logger.addHandler(file_handler)
    apply_default_formatter(file_handler)
    apply_default_formatter(console_handler)

    if agent_id:
        logger = logging.LoggerAdapter(logger, {'agent_id': str(agent_id)})
    return logger


def get_file_logger() -> logging.FileHandler:
    os.makedirs(os.path.dirname(settings.get().LOG_FILE_PATH), exist_ok=True)
    file_handler = logging.FileHandler(settings.get().LOG_FILE_PATH)
    file_handler.setLevel(logging.DEBUG)
    return file_handler


def get_console_logger() -> logging.StreamHandler:
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)
    return console_handler


def apply_default_formatter(handler: logging.Handler):
    formatter = jsonlogger.JsonFormatter(LOGGING_MESSAGE_FORMAT)
    handler.setFormatter(formatter)



from typing import List
from typing import Dict
from typing import Union
import os

from bs4 import BeautifulSoup
from langchain.document_loaders import BSHTMLLoader
from langchain.schema import Document


def execute(paths: List[str]) -> List[Document]:
    documents: List[Document] = []
    for path in paths:
        if os.path.isdir(path):
            scraped_documents= _scrape_folder(path)
        else:
            scraped_documents = _scrape_html(path)
        documents.extend(scraped_documents)
    return documents


def _scrape_folder(folder: str) -> List[Document]:
    documents = []
    files = []
    for dir_path, dir_names, file_names in os.walk(folder):
        for file_name in file_names:
            if os.path.splitext(file_name)[1] in [".html", ".htm"]:
                files.append(os.path.join(dir_path, file_name))
    for file in files:
        documents.extend(_scrape_html(file))
    return documents


def _scrape_html(filename) -> List[Document]:
    try:
        loader = CustomHtmlLoader(filename)
        return loader.load()
    except:
        return []


class CustomHtmlLoader(BSHTMLLoader):

    def load(self) -> List[Document]:
        with open(self.file_path, "r", encoding=self.open_encoding) as f:
            soup = BeautifulSoup(f, **self.bs_kwargs)

        section = soup.find(class_="section")
        navigation = soup.find_all(class_="nextprev")
        breadcrumbs = soup.find_all(class_="breadcrumbs")
        for breadcrumb in breadcrumbs:
            breadcrumb.extract()
        for child in navigation:
            child.extract()
        text = section.get_text(self.get_text_separator)

        if soup.title:
            title = str(soup.title.string)
        else:
            title = ""

        metadata: Dict[str, Union[str, None]] = {
            "source": self.file_path,
            "title": title,
        }
        return [Document(page_content=text, metadata=metadata)]



import json

import requests

from opencopilot import settings
from opencopilot.logger import api_logger

HEADERS = {
    "Content-type": "application/json"
}

logger = api_logger.get()


def post_error(error_type: str, error: str):
    _post_message(error_type, error)


def _post_message(error_description: str, error: str):
    url = settings.get().SLACK_WEBHOOK
    if url is None or url == "":
        logger.error(f"Slack notification would be sent: {error_description} - {error}")
        return
    try:
        data = _get_data(error_description, error)
        requests.post(url, headers=HEADERS, data=json.dumps(data))
    except:
        logger.error("Failed to post error message on slack", exc_info=1)


def _get_data(error_description: str, error: str):
    return {
        "blocks": [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "*{}*\n\nError: {}".format(
                        error_description, error)
                }
            }
        ]
    }



import json
import os
from typing import Callable
from typing import List

from opencopilot import settings
from langchain.schema import Document
from opencopilot.utils.loaders import fullstack_deeplearning_loader_use_case, \
    openai_docs_loader_use_case, url_loader_use_case, openai_api_docs_loader_use_case, \
    prompting_guide_docs_loader_use_case, html_loader_use_case, blog_loader_use_case, \
    anthropic_docs_loader_use_case, gitbook_loader_use_case

LOADERS_MAP = {
    "gitbook": gitbook_loader_use_case.execute,
    "url": url_loader_use_case.execute,
    "blog": blog_loader_use_case.execute,
    "html": html_loader_use_case.execute,
    "openai_docs": openai_docs_loader_use_case.execute,
    "openai_api_docs": openai_api_docs_loader_use_case.execute,
    "anthropic_docs": anthropic_docs_loader_use_case.execute,
    "prompting_guide_docs": prompting_guide_docs_loader_use_case.execute,
    "fullstack_deeplearning": fullstack_deeplearning_loader_use_case.execute,
}


def execute():
    if settings.get().copilot_config and settings.get().copilot_config.data.loaders:
        for data_group in settings.get().copilot_config.data.loaders:
            if loader := LOADERS_MAP.get(data_group):
                configuration = settings.get().copilot_config.data.loaders[data_group]
                _scrape(
                    os.path.join(
                        settings.get().DATA_DIR,
                        configuration["output"],
                    ),
                    loader,
                    configuration["sources"],
                    configuration.get("base_url"),
                )


def _scrape(
        file_dump_path: str,
        loader_function: Callable[[List[str]], List[Document]],
        urls: List[str],
        base_url: str = None,
) -> None:
    if not os.path.exists(file_dump_path):
        scraped_documents = loader_function(urls)
        if base_url:
            for doc in scraped_documents:
                doc.metadata["source"] = base_url + os.path.basename(doc.metadata['source'])

        json_docs = [d.json() for d in scraped_documents]
        formatted_docs = [json.loads(d) for d in json_docs]

        os.makedirs(settings.get().DATA_DIR, exist_ok=True)
        with open(file_dump_path, "w") as f:
            f.write(json.dumps(formatted_docs, indent=4))



import pickle
from typing import List

import xxhash
from langchain.embeddings import OpenAIEmbeddings

from opencopilot import settings
from opencopilot.logger import api_logger

logger = api_logger.get()


class CachedOpenAIEmbeddings(OpenAIEmbeddings):
    use_local_cache: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        object.__setattr__(self, "_cache", {})
        object.__setattr__(self, "_embeddings_cache_filename", settings.get().COPILOT_NAME + "_embeddings_cache.pkl")
        self._load_local_cache()

    def embed_documents(self, texts: List[str], **kwargs) -> List[List[float]]:
        if self.use_local_cache:
            return self._embed_documents_cached(texts)
        return super().embed_documents(texts)

    def embed_query(self, text: str) -> List[float]:
        return super().embed_query(text)

    def _embed_documents_cached(self, texts: List[str]) -> List[List[float]]:
        embeddings = []
        for text in texts:
            text_hash = self._hash(text)
            # pylint: disable-next=no-member
            if embedding := self._cache.get(text_hash):
                embeddings.append(embedding)
            else:
                embedding = super().embed_documents([text])[0]
                # pylint: disable-next=no-member
                self._cache[text_hash] = embedding
                embeddings.append(embedding)
        return embeddings

    def _hash(self, text) -> str:
        return xxhash.xxh64(text.encode("utf-8")).hexdigest()

    def _load_local_cache(self):
        try:
            # pylint: disable-next=no-member
            with open(self._embeddings_cache_filename, "rb") as f:
                data = pickle.load(f)
                object.__setattr__(self, "_cache", data)
        except:
            pass

    def save_local_cache(self):
        if self.use_local_cache:
            try:
                with open(self._embeddings_cache_filename, "wb") as f:
                    pickle.dump(self._cache, f)
            except Exception as e:
                logger.warning(f"Failed to save embeddings cache to {self._embeddings_cache_filename}")


def execute(use_local_cache: bool = False):
    openai_api_base = None
    headers = None
    if settings.get().HELICONE_API_KEY:
        openai_api_base = settings.get().HELICONE_BASE_URL
        headers = {
            "Helicone-Auth": "Bearer " + settings.get().HELICONE_API_KEY,
            "Helicone-Cache-Enabled": "true",
        }
    return CachedOpenAIEmbeddings(
        disallowed_special=(), use_local_cache=use_local_cache,
        openai_api_base=openai_api_base,
        headers=headers
    )



from opencopilot.domain.chat.entities import ChatContextInput
from opencopilot.domain.chat.entities import ChatContextOutput
from opencopilot.repository.conversation_user_context_repository import ConversationUserContextRepositoryLocal


def execute(
        data_input: ChatContextInput,
        repository: ConversationUserContextRepositoryLocal
) -> ChatContextOutput:
    repository.save_context(data_input)
    return ChatContextOutput(
        response="OK"
    )



async def add_response_headers(response, duration):
    if duration:
        response.headers["Response-Time"] = str(duration)
    response.headers["Access-Control-Allow-Origin"] = "*"
    response.headers["Access-Control-Allow-Headers"] = \
        "Content-Type,Authorization,Response-Time,source"
    response.headers["Access-Control-Allow-Methods"] = \
        "GET,PUT,POST,DELETE,OPTIONS"
    return response



import os
from datetime import timedelta
from typing import Callable
from typing import Literal
from typing import Optional

import uvicorn
from langchain.schema import Document

from . import settings
from .settings import Settings


class OpenCopilot:

    def __init__(
            self,
            openai_api_key: Optional[str] = None,
            copilot_name: str = "default",
            api_base_url: str = "http://127.0.0.1/",
            api_port: int = 3000,
            environment: str = "local",
            allowed_origins: str = "*",
            application_name: str = "backend-service",
            log_file_path="../logs/logs-backend-service.log",
            weaviate_url: str = "http://localhost:8080/",
            weaviate_read_timeout: int = 120,
            llm_model_name: Literal["gpt-3.5-turbo-16k", "gpt-4"] = "gpt-4",
            max_document_size_mb: int = 50,
            slack_webhook: str = "",
            auth_type: Optional[str] = None,
            api_key: str = "",
            jwt_client_id: str = "",
            jwt_client_secret: str = "",
            jwt_token_expiration_seconds: int = timedelta(days=1).total_seconds(),
            helicone_api_key: str = "",
            helicone_rate_limit_policy: str = "3;w=60;s=user",
    ):
        if not openai_api_key:
            openai_api_key = os.getenv("OPENAI_API_KEY")
        assert openai_api_key, "OPENAI_API_KEY must be passed to OpenCopilot or be set in the environment."

        settings.set(
            Settings(
                OPENAI_API_KEY=openai_api_key,
                COPILOT_NAME=copilot_name,
                API_PORT=api_port,
                API_BASE_URL=api_base_url,
                ENVIRONMENT=environment,
                ALLOWED_ORIGINS=allowed_origins,
                APPLICATION_NAME=application_name,
                LOG_FILE_PATH=log_file_path,
                WEAVIATE_URL=weaviate_url,
                WEAVIATE_READ_TIMEOUT=weaviate_read_timeout,
                MODEL=llm_model_name,
                MAX_DOCUMENT_SIZE_MB=max_document_size_mb,
                SLACK_WEBHOOK=slack_webhook,
                AUTH_TYPE=auth_type,
                API_KEY=api_key,
                JWT_CLIENT_ID=jwt_client_id,
                JWT_CLIENT_SECRET=jwt_client_secret,
                JWT_TOKEN_EXPIRATION_SECONDS=jwt_token_expiration_seconds,
                HELICONE_API_KEY=helicone_api_key,
                HELICONE_RATE_LIMIT_POLICY=helicone_rate_limit_policy
            ))

        self.api_port = api_port
        self.data_loaders = []
        self.local_files_dirs = []
        self.local_file_paths = []
        self.documents = []

    def __call__(self, *args, **kwargs):
        from .repository.documents import document_loader
        from .repository.documents import document_store
        from opencopilot.repository.documents.document_store import WeaviateDocumentStore
        from opencopilot.repository.documents.document_store import EmptyDocumentStore
        if self.data_loaders or self.local_files_dirs or self.local_file_paths:
            self.document_store = WeaviateDocumentStore()
        else:
            self.document_store = EmptyDocumentStore()
        document_store.init_document_store(self.document_store)

        for data_loader in self.data_loaders:
            self.documents.extend(data_loader())

        for data_dir in self.local_files_dirs:
            text_splitter = self.document_store.get_text_splitter()
            self.documents.extend(document_loader.execute(data_dir, False, text_splitter))

        if self.documents:
            self.document_store.ingest_data(self.documents)

        from .app import app
        uvicorn.run(app, port=self.api_port)

    @staticmethod
    def add_prompt(prompt_file: str) -> None:
        settings.init_prompt_file_location(prompt_file)

    def data_loader(
            self,
            function: Callable[[], Document]
    ):
        self.data_loaders.append(function)

    def add_local_files_dir(self, files_dir: str) -> None:
        self.local_files_dirs.append(files_dir)

    # def add_local_file(self, file_path: str) -> None:
    #    self.local_file_paths.append(file_path)



from datetime import datetime

from opencopilot.logger import api_logger
from opencopilot.domain.chat import validate_urls_use_case
from opencopilot.domain.chat.entities import MessageModel
from opencopilot.domain.chat.entities import UserMessageInput
from opencopilot.domain.chat.results import get_gpt_result_use_case
from opencopilot.domain.chat.utils import get_system_message
from opencopilot.repository.conversation_history_repository import ConversationHistoryRepositoryLocal
from opencopilot.repository.conversation_logs_repository import ConversationLogsRepositoryLocal
from opencopilot.repository.documents.document_store import DocumentStore

logger = api_logger.get()


async def execute(
        domain_input: UserMessageInput,
        document_store: DocumentStore,
        history_repository: ConversationHistoryRepositoryLocal,
        logs_repository: ConversationLogsRepositoryLocal,
) -> MessageModel:
    system_message = get_system_message()
    context = []
    if "{context}" in system_message:
        context = document_store.find(domain_input.message)
    message_timestamp = datetime.now().timestamp()
    result = await get_gpt_result_use_case.execute(
        domain_input,
        system_message,
        context,
        logs_repository=logs_repository,
        history_repository=history_repository,
    )

    validate_urls_use_case.execute(result, domain_input.chat_id)

    response_timestamp = datetime.now().timestamp()

    history_repository.save_history(
        domain_input.message,
        result,
        message_timestamp,
        response_timestamp,
        domain_input.chat_id,
        domain_input.response_message_id,
    )
    sources = [document.metadata.get('source') for document in context]

    return MessageModel(
        chat_id=domain_input.chat_id,
        content=result,
        sources=sources
    )



import os
import json
import time
import re
import urllib.request
import argparse
from typing import List, Dict

API_BASE_URL = os.getenv("API_BASE_URL", "https://api.unity.sidekik.ai/")
UNITY_LOG_FILE_PATH = os.path.expanduser('~/Library/Logs/Unity/Editor.log')
LOG_LINES_TO_INCLUDE = 20


def _send_info(conversation_id: str, info: Dict) -> bool:
    try:
        data = {
            "context": json.dumps(info) 
        }
        data_json = json.dumps(data)
        data_bytes = data_json.encode('utf-8')  # convert string to bytes
        req = urllib.request.Request(f"{API_BASE_URL}v0/conversation/{conversation_id}/context")
        req.add_header('Content-Type', 'application/json; charset=utf-8')
        req.add_header('Content-Length', len(data_bytes))
        response = urllib.request.urlopen(req, data_bytes)
        response_json = json.loads(response.read())
        return response_json["response"] == "OK"
    except Exception as e:
        print(f"Error sending information to Unity Copilot: {str(e)}")
        return False


def _walk_folder_tree(directory, abs_path):
    all_files_folders = []
    for root, dirs, files in os.walk(os.path.join(abs_path, directory)):
        relative_root = os.path.relpath(root, abs_path)
        for name in files:
            all_files_folders.append(os.path.join(relative_root, name))
        for name in dirs:
            all_files_folders.append(os.path.join(relative_root, name))

    return all_files_folders


def _get_packages(project_path: str):
    try:
        with open(f"{project_path}/Packages/packages-lock.json", "r") as f:
            return json.load(f)
    except:
        return {}


def _get_directory_tree(project_path):
    directories = ["Assets", "ProjectSettings", "Packages"]

    tree = []
    for directory in directories:
        tree.extend(_walk_folder_tree(directory, project_path))
    return tree


def _extract_info(line):
    pattern = r'\w+:\s+(?:Info\s+)?(?P<Info>.*)'
    info = re.search(pattern, line)
    return info.group('Info') if info else None


def _parse_os_unity_info(data: List[str]) -> Dict:
    '''
        Unity Editor version:    2021.3.28f1 (232e59c3f087)
        Branch:                  2021.3/release
        Build type:              Release
        Batch mode:              NO
        macOS version:           Version 13.2 (Build 22D49)
        Darwin version:          22.3.0
        Architecture:            arm64
        Running under Rosetta:   NO
        Available memory:        16384 MB
    '''

    return {
        "unity_version": _extract_info(data[0]),
        "macOS_version": _extract_info(data[4]),
        "darwin version": _extract_info(data[5]),
        "architecture": _extract_info(data[6]),
        "rosetta": _extract_info(data[7]) == "YES",
        "available_memory_mb": int(_extract_info(data[8]).split()[0]) 
    }


def _process_log_file() -> Dict:
    try:
        with open(UNITY_LOG_FILE_PATH, 'r') as file:
            lines = file.readlines()
            # extract OS and Unity information
            info = _parse_os_unity_info(lines[:9])
            # scan the log to find project information
            project_path = None
            for i, line in enumerate(lines):
                if '-projectpath' in line or "-createproject" in line:
                    project_path = re.findall(r'(/[\w\W]*)', lines[i+1])
                    break
            if project_path:
                project_path = project_path[0].strip()
                packages = _get_packages(project_path)
                project_structure = _get_directory_tree(project_path)
                # put everything together
                info["project_path"] = project_path
                info["project_structure"] = project_structure
                info["packages"] = packages
            info["editor_log"] = lines[-LOG_LINES_TO_INCLUDE:]
            return info
    except FileNotFoundError:
        print(f"{UNITY_LOG_FILE_PATH} does not exist.")
    except PermissionError:
        print(f"Permission denied for {UNITY_LOG_FILE_PATH}.")
    except Exception as e:
        print(f"An error occurred: {str(e)}")


def _monitor_unity_log_file() -> Dict:
    last_log_mt = None
    last_package_mt = None
    current_package_mt = None
    packages_file = None

    while True:
        try:
            # Check if the file has been modified
            current_log_mt = os.path.getmtime(UNITY_LOG_FILE_PATH)
            if packages_file:
                current_package_mt = os.path.getmtime(packages_file)
            if last_log_mt != current_log_mt or last_package_mt != current_package_mt:
                # modified, get info
                last_log_mt = current_log_mt
                last_package_mt = current_package_mt
                info = _process_log_file()
                if "project_path" in info:
                    packages_file = f"{info['project_path']}/Packages/packages-lock.json"
                yield info
        except FileNotFoundError:
            print(f"{UNITY_LOG_FILE_PATH} does not exist.")
            return
        except PermissionError:
            print(f"Permission denied for {UNITY_LOG_FILE_PATH}.")
            return
        except Exception as e:
            print(f"An error occurred: {str(e)}")
            return

        # Pause for a bit
        time.sleep(2)


def main(conversation_id: str):
    for info in _monitor_unity_log_file():
        _send_info(conversation_id, info)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "conversation_id",
        type=str
    )
    args = parser.parse_args()
    main(args.conversation_id)



from typing import Dict
from typing import List
from typing import Optional
from uuid import UUID

from opencopilot.domain.debug.entities import MessageDebugResult
from opencopilot.domain.debug.entities import TextWithTokens
from opencopilot.repository.conversation_history_repository import ConversationHistoryRepositoryLocal
from opencopilot.repository.conversation_logs_repository import ConversationLogsRepositoryLocal


def execute(
        conversation_id: UUID,
        message_id: str,
        history_repository: ConversationHistoryRepositoryLocal,
        logs_repository: ConversationLogsRepositoryLocal,
) -> MessageDebugResult:
    history = history_repository.get_history(conversation_id)
    logs_history = logs_repository.get_logs_by_message(
        conversation_id, message_id)
    return MessageDebugResult(
        prompt_template=_get_logs_history_value(logs_history, "prompt_template"),
        data_sources=None,
        user_question=_get_history_value(history, message_id, "prompt"),
        context=_get_logs_history_value(logs_history, "context"),
        chat_history=_get_logs_history_value(logs_history, "history"),
        full_prompt=_get_logs_history_value(logs_history, "prompt_text"),
        llm_response=_get_history_value(history, message_id, "response"),
    )


def _get_history_value(
        history: List[Dict],
        message_id: str,
        value_key: str,
) -> Optional[TextWithTokens]:
    for h in history:
        if h.get("response_message_id") == message_id:
            return TextWithTokens(
                text=h.get(value_key),
                token_count=None,
            )

    return None


def _get_logs_history_value(
        history: List[Dict],
        value_key: str,
) -> Optional[TextWithTokens]:
    for h in history:
        if value := h.get(value_key):
            return TextWithTokens(
                text=value,
                token_count=h.get("token_count"),
            )
    return None



from opencopilot.domain.chat import chat_context_use_case
from opencopilot.domain.chat.entities import ChatContextInput
from opencopilot.repository.conversation_user_context_repository import \
    ConversationUserContextRepositoryLocal
from opencopilot.service.chat.entities import ChatContextRequest
from opencopilot.service.entities import ApiResponse
from opencopilot.service.utils import get_uuid


def execute(
        conversation_id: str,
        request: ChatContextRequest,
        repository: ConversationUserContextRepositoryLocal
) -> ApiResponse:
    conversation_id = get_uuid(conversation_id, "conversation_id")
    response = chat_context_use_case.execute(
        data_input=ChatContextInput(
            conversation_id=conversation_id,
            context=request.context,
        ),
        repository=repository
    )
    return ApiResponse(
        response=response.response
    )



import os
from dataclasses import dataclass
from uuid import UUID

from opencopilot import settings
from opencopilot.repository.conversation_history_repository import ConversationHistoryRepositoryLocal


@dataclass(frozen=True)
class History:
    template_with_history: str
    formatted_history: str


def add_history(
        template: str,
        chat_id: UUID,
        history_repository: ConversationHistoryRepositoryLocal,
) -> History:
    os.makedirs(settings.get().CONVERSATIONS_DIR, exist_ok=True)
    history = history_repository.get_prompt_history(
        chat_id, settings.get().PROMPT_HISTORY_INCLUDED_COUNT)
    history = history.replace("{", "{{").replace("}", "}}")
    return History(
        template_with_history=template.replace("{history}", history, 1),
        formatted_history=history,
    )


def get_system_message() -> str:
    try:
        with open(settings.get().PROMPT_FILE, "r") as f:
            return f.read()
    except:
        return settings.get().DEFAULT_PROMPT



from typing import List
from typing import Optional
from urllib.parse import urljoin

from bs4 import BeautifulSoup
from langchain.schema import Document
from playwright.sync_api import sync_playwright


def execute(urls: List[str]) -> List[Document]:
    docs: List[Document] = []
    for url in urls:
        soup = _scrape_html_with_playwright(url)
        sections = soup.find_all(class_="section")
        for section in sections:
            urls = section.find_all(class_="anchor-heading-link")
            if len(urls) and urls[0].attrs.get("href"):
                href = urls[0].attrs.get("href")
                docs.append(Document(
                    page_content=section.text,
                    metadata={"source": urljoin(url, href)}
                ))
    return docs


def _scrape_html_with_playwright(url) -> Optional[BeautifulSoup]:
    soup = None
    try:
        with sync_playwright() as p:
            browser = p.chromium.launch(headless=True)
            try:
                page = browser.new_page()
                page.goto(url)
                page_source = page.content()
                soup = BeautifulSoup(page_source, features="lxml")
            except:
                pass
            browser.close()
    except Exception as e:
        print(e)
    return soup



from typing import List

import requests
from bs4 import BeautifulSoup

from langchain.schema import Document

from opencopilot.utils.loaders import url_loader_use_case


def execute(blog_urls: List[str]) -> List[Document]:
    documents: List[Document] = []
    for blog_url in blog_urls:
        urls = _get_urls(blog_url)
        documents.extend(url_loader_use_case.execute(urls))
    return documents


def _get_urls(blog_url: str) -> List[str]:
    session = requests.Session()
    html_doc = session.get(blog_url, timeout=10)
    html_doc.encoding = html_doc.apparent_encoding
    soup = BeautifulSoup(html_doc.text, "html.parser")
    links = soup.find_all("a")

    urls: List[str] = []
    for link in links:
        # Check if url points to blog and primitive date check
        if "/blog/" in link.attrs.get("href") and ("2022" in link.text or "2023" in link.text):
            urls.append(_get_formatted_url(link.attrs.get("href")))
    return urls


def _get_formatted_url(path: str) -> str:
    if path.startswith("https://"):
        return path
    return "https://readyplayerdev.me" + path



from typing import List

from langchain.document_loaders import PlaywrightURLLoader
from langchain.document_loaders import UnstructuredURLLoader
from langchain.schema import Document


def execute(urls: List[str]) -> List[Document]:
    documents: List[Document] = []
    for url in urls:
        scraped_documents = _scrape_webpage(url)
        documents.extend(scraped_documents)
    return documents


def _scrape_webpage(url) -> List[Document]:
    documents = _scrape_html(url)
    if not documents or len(documents) == 1 and len(documents[0].page_content) < 20:
        documents = _scrape_html_with_playwright(url)
    return documents


def _scrape_html(url) -> List[Document]:
    try:
        loader = UnstructuredURLLoader(urls=[url])
        return loader.load()
    except:
        return []


def _scrape_html_with_playwright(url) -> List[Document]:
    try:
        loader = PlaywrightURLLoader(
            urls=[url],
            remove_selectors=["header", "footer"]
        )
        return loader.load()
    except Exception as e:
        print(e)
    return []



import json
from dataclasses import dataclass
from pathlib import Path
from typing import List


@dataclass(frozen=True)
class UserFeedbackAnalytics:
    correctness: List[int]
    helpfulness: List[int]
    easy_to_understand: List[int]


def execute(conversations_dir="conversations") -> UserFeedbackAnalytics:
    feedbacks = _collect_feedbacks(conversations_dir)
    print("\nRaw feedbacks:")
    print(feedbacks)

    correctness = []
    helpfulness = []
    easy_to_understand = []
    for chat_feedbacks in feedbacks:
        for f in chat_feedbacks:
            if f.get("correctness"):
                correctness.append(f.get("correctness"))
            if f.get("helpfulness"):
                helpfulness.append(f.get("helpfulness"))
            if f.get("easy_to_understand"):
                easy_to_understand.append(f.get("easy_to_understand"))

    print("\nMetrics:")
    _print_metric("correctness", correctness)
    _print_metric("helpfulness", helpfulness)
    _print_metric("easy_to_understand", easy_to_understand)
    total_feedbacks = len(correctness) + len(helpfulness) + len(easy_to_understand)
    print(f"\n  total feedbacks given:", total_feedbacks)
    return UserFeedbackAnalytics(
        correctness=correctness,
        helpfulness=helpfulness,
        easy_to_understand=easy_to_understand
    )


def _print_metric(label: str, metric: List[int]) -> None:
    print(f"  {label}:")
    print(f"    len:", len(metric))
    print(f"    min:", min(metric))
    print(f"    max:", max(metric))
    print(f"    avg:", sum(metric) / len(metric))


def _collect_feedbacks(conversations_dir: str):
    pathlist = Path(conversations_dir).rglob('*.json')
    result = []
    for path in pathlist:
        feedbacks = _get_feedbacks_from_file(str(path))
        if feedbacks:
            result.append(feedbacks)
    return result


def _get_feedbacks_from_file(file_path: str):
    result = []
    data = _read_file(file_path)
    data = json.loads(data)
    for item in data:
        feedback = item.get("user_feedback")
        if feedback:
            result.append(feedback)
    return result


def _read_file(file_path: str) -> str:
    with open(file_path, "r", encoding="utf-8") as file:
        return file.read()


if __name__ == '__main__':
    execute()



from opencopilot.domain.chat import get_chat_history_use_case
from opencopilot.repository.conversation_history_repository import ConversationHistoryRepositoryLocal
from opencopilot.service.chat.entities import ChatHistoryRequest
from opencopilot.service.chat.entities import ChatHistoryResponse
from opencopilot.service.utils import get_uuid


async def execute(
        request: ChatHistoryRequest,
        history_repository: ConversationHistoryRepositoryLocal,
) -> ChatHistoryResponse:
    chat_id = get_uuid(request.chat_id, "chat_id")
    messages = await get_chat_history_use_case.execute(
        chat_id,
        history_repository,
    )
    return ChatHistoryResponse(
        response="OK",
        chat_id=str(chat_id),
        messages=messages,
    )



from typing import List, Optional

import tqdm
import weaviate
from langchain.schema import Document
from langchain.text_splitter import CharacterTextSplitter
from langchain.text_splitter import TextSplitter
from langchain.vectorstores import Weaviate

from opencopilot import settings
from opencopilot.logger import api_logger
from opencopilot.utils import get_embedding_model_use_case
from opencopilot.utils.get_embedding_model_use_case import CachedOpenAIEmbeddings

logger = api_logger.get()


class DocumentStore:
    document_embed_model = "text-embedding-ada-002"
    document_chunk_size = 2000

    def get_embeddings_model(self) -> CachedOpenAIEmbeddings:
        return get_embedding_model_use_case.execute(use_local_cache=True)

    def get_text_splitter(self) -> TextSplitter:
        return CharacterTextSplitter.from_tiktoken_encoder(
            chunk_size=self.document_chunk_size,
            model_name=self.document_embed_model,
            separator=" ",
            disallowed_special=(),
        )

    def ingest_data(self, documents: List[Document]):
        pass

    def find(self, query: str, **kwargs) -> List[Document]:
        return []


class WeaviateDocumentStore(DocumentStore):
    ingest_batch_size = 100

    weaviate_index_name = "LangChain"  # TODO: Weaviate specific?

    def __init__(self):
        self.documents = []
        self.embeddings = self.get_embeddings_model()
        self.weaviate_client = self._get_weaviate_client()
        self.vector_store = self._get_vector_store()

    def _get_weaviate_client(self):
        return weaviate.Client(
            url=settings.get().WEAVIATE_URL,
            timeout_config=(10, settings.get().WEAVIATE_READ_TIMEOUT)
        )

    def _get_vector_store(self):
        metadatas = [d.metadata for d in self.documents]
        attributes = list(metadatas[0].keys()) if metadatas else None
        return Weaviate(
            self.weaviate_client,
            index_name=self.weaviate_index_name,
            text_key="text",
            embedding=self.embeddings,
            attributes=attributes,
            by_text=False
        )

    def ingest_data(self, documents: List[Document]):
        self.documents = documents
        batch_size = self.ingest_batch_size
        print(f"Got {len(documents)} documents, embedding with batch size: {batch_size}")
        self.weaviate_client.schema.delete_all()

        for i in tqdm.tqdm(range(0, int(len(documents) / batch_size) + 1), desc="Embedding.."):
            batch = documents[i * batch_size: (i + 1) * batch_size]
            self.vector_store.add_documents(batch)

        self.embeddings.save_local_cache()
        self.vector_store = self._get_vector_store()

    def find(self, query: str, **kwargs) -> List[Document]:
        k = kwargs.get("k") or settings.get().MAX_CONTEXT_DOCUMENTS_COUNT
        documents = self.vector_store.similarity_search(
            query,
            k=k
        )
        return documents[:k]


class EmptyDocumentStore(DocumentStore):
    pass


DOCUMENT_STORE = Optional[DocumentStore]


def init_document_store(document_store: DocumentStore):
    global DOCUMENT_STORE
    DOCUMENT_STORE = document_store


def get_document_store() -> DocumentStore:
    return DOCUMENT_STORE



import json
from dataclasses import dataclass
from pathlib import Path
from typing import List


@dataclass(frozen=True)
class UserFeedbackResult:
    conversation_id: str
    prompt: str
    response: str
    correctness: int
    helpfulness: int
    easy_to_understand: int
    free_form_feedback: str


def execute(conversations_dir="conversations") -> List[UserFeedbackResult]:
    result = []
    pathlist = Path(conversations_dir).rglob('*.json')
    for path in pathlist:
        feedbacks = _get_feedbacks_from_file(str(path))
        if feedbacks:
            result += feedbacks
    return result


def _get_feedbacks_from_file(file_path: str):
    result = []
    data = _read_file(file_path)
    data = json.loads(data)
    for item in data:
        feedback = item.get("user_feedback")
        if feedback:
            result.append(UserFeedbackResult(
                conversation_id=file_path.split("/")[-1].replace(".json", ""),
                prompt=item.get("prompt"),
                response=item.get("response"),
                correctness=feedback.get("correctness"),
                helpfulness=feedback.get("helpfulness"),
                easy_to_understand=feedback.get("easy_to_understand"),
                free_form_feedback=feedback.get("free_form_feedback"),
            ))
    return result


def _read_file(file_path: str) -> str:
    with open(file_path, "r", encoding="utf-8") as file:
        return file.read()


if __name__ == '__main__':
    all_feedbacks = execute()
    for f in all_feedbacks:
        print("\n\nFeedback:")
        print(f)



import json
import os
import uuid
from pathlib import Path
from typing import Optional

import requests
from dotenv import load_dotenv

headers = {
    "accept": "application/json",
    "Content-Type": "application/json"
}
DEFAULT_MESSAGE = "Hi"


def _get_stream(url: str, message: str = DEFAULT_MESSAGE, jwt_token: str = None):
    data = {
        "inputs": message
    }
    if jwt_token:
        headers["Authorization"] = "Bearer " + jwt_token
    s = requests.Session()
    with s.post(url, headers=headers, json=data, stream=True) as resp:
        for line in resp.iter_lines():
            if line:
                yield line


def _process_text(line):
    line = line.decode("utf-8")
    try:
        line = json.loads(line)
        if error := line.get("error"):
            print("ERROR:", error)
            raise Exception("Error in stream")
        return line["text"]
    except:
        return ""


def conversation(
        base_url: str,
        conversation_id: uuid.UUID,
        message: str = DEFAULT_MESSAGE,
):
    jwt_token = _get_jwt_token(base_url)
    if jwt_token:
        headers["Authorization"] = "Bearer " + jwt_token
    url = f"{base_url}/v0/conversation/{conversation_id}"
    data = {
        "inputs": message
    }
    return requests.post(url, json=data, headers=headers)


def conversation_stream(
        base_url: str,
        conversation_id: uuid.UUID,
        message: str = DEFAULT_MESSAGE,
        stream: bool = False
):
    jwt_token = _get_jwt_token(base_url)
    url = f"{base_url}/v0/conversation_stream/{conversation_id}"
    output = ""
    for text in _get_stream(url, message=message, jwt_token=jwt_token):
        text = _process_text(text)
        output += text
        if stream:
            print(text, end="", flush=True)
    return output


def _get_jwt_token(base_url: str) -> Optional[str]:
    env_path = Path(".") / ".env"
    load_dotenv(dotenv_path=env_path)
    jwt_client_id = os.getenv("JWT_CLIENT_ID", "")
    jwt_client_secret = os.getenv("JWT_CLIENT_SECRET", "")
    url = f"{base_url}/v0/token"
    data = {
        "client_id": jwt_client_id,
        "client_secret": jwt_client_secret,
        "user_id": "test@local.host"
    }
    try:
        token_result = requests.post(url, headers=headers, json=data)
        result_json = token_result.json()
        return result_json["token"]
    except:
        return None


if __name__ == '__main__':
    _result = conversation_stream(
        # TODO: fix base_url
        base_url=f"http://0.0.0.0:3000",
        conversation_id=uuid.uuid4()
    )
    print("result:", _result)



from enum import Enum


class RequestStateKey(str, Enum):
    """
    Keys for using request store to transfer information between http middlewares
    """
    REQUEST_ID = "request_id"
    REQUEST_SOURCE = "request_source"
    REQUEST_ORIGIN = "request_origin"
    REQUEST_USER_AGENT = "request_user_agent"
    IP_ADDRESS = "ip_address"
    COUNTRY = "country"



from typing import Optional

from fastapi import APIRouter, Header
from fastapi import Body
from fastapi import Depends
from fastapi import Path
from fastapi.responses import StreamingResponse
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from opencopilot.logger import api_logger
from opencopilot.authorization import validate_api_key_use_case
from opencopilot.repository.conversation_history_repository import \
    ConversationHistoryRepositoryLocal
from opencopilot.repository.conversation_logs_repository import ConversationLogsRepositoryLocal
from opencopilot.repository.conversation_user_context_repository import \
    ConversationUserContextRepositoryLocal
from opencopilot.repository.documents import document_store
from opencopilot.routers import routing_utils
from opencopilot.service import utils
from opencopilot.service.chat import chat_context_service, chat_service
from opencopilot.service.chat import chat_feedback_service, \
    chat_streaming_service, \
    chat_history_service
from opencopilot.service.chat.entities import ChatContextRequest
from opencopilot.service.chat.entities import ChatFeedbackRequest
from opencopilot.service.chat.entities import ChatHistoryRequest, ChatHistoryResponse
from opencopilot.service.chat.entities import ChatRequest
from opencopilot.service.chat.entities import ChatResponse
from opencopilot.service.entities import ApiResponse

TAG = "Chat"
router = APIRouter()
router.openapi_tags = [TAG]
router.title = "Chat router"

logger = api_logger.get()

STREAM_RESPONSE_DESCRIPTION = """
A stream of objects, delimited by newlines. Each object will be of the following form:
```
{
    "text": "some text" # the next chunk of the message from the copilot
    "error": ""         # if present, a string description of the error that occurred
}
```

For example, the message "I like to eat apples" might be streamed as follows:

```
{"text": "I like"}
{"text": " to eat"}
{"text": " apples"}
```
"""


class ConversationInput(BaseModel):
    inputs: str = Field(
        ...,
        description="Message to be answered by LLM.",
        example="How do I make a delicious lemon cheesecake?",
    )
    response_message_id: Optional[str] = Field(
        description="Optional message ID UUID, that can be used to get debug info.",
        example="e0f07564-b26b-4074-876e-d47b92ef767a",
    )

    @validator("response_message_id")
    # pylint: disable-next=no-self-argument
    def id_check(cls, v, values):
        if v:
            utils.get_uuid(v, "response_message_id")
        return v

    class Config:
        schema_extra = {
            "example": {
                "inputs": "How do I make a delicious lemon cheesecake?",
            }
        }


@router.post(
    "/conversation/{conversation_id}",
    summary="Send message to the base copilot.",
    tags=[TAG]
)
async def handle_conversation(
        email: Optional[str] = Header(default=None),
        conversation_id: str = Path(...,
                                    description="The ID of the conversation. To start a new conversation, you should pass in a random uuid (Python: `import uuid; uuid.uuid4()`). To continue a conversation, re-use the same uuid."),
        payload: ConversationInput = Body(...,
                                          description="Input and parameters for the conversation."),
        user_id: str = Depends(validate_api_key_use_case.execute)
):
    request = ChatRequest(
        chat_id=conversation_id,
        message=payload.inputs,
        response_message_id=payload.response_message_id,
        email=user_id or email
    )

    history_repository = ConversationHistoryRepositoryLocal()
    logs_repository = ConversationLogsRepositoryLocal()

    response: ChatResponse = await chat_service.execute(
        request,
        document_store.get_document_store(),
        history_repository,
        logs_repository,
    )
    return routing_utils.to_json_response({
        "generated_text": response.message,
        "sources": response.sources
    })


@router.post(
    "/conversation/{conversation_id}/feedback",
    tags=[TAG],
    summary="Send feedback to a conversation.",
    response_model=ApiResponse,
)
async def post_feedback(
        conversation_id: str = Path(..., description="The ID of the conversation."),
        payload: ChatFeedbackRequest = Body(..., description="User feedback")
):
    history_repository = ConversationHistoryRepositoryLocal()
    response = chat_feedback_service.execute(
        conversation_id=conversation_id,
        request=payload,
        repository=history_repository)
    return routing_utils.to_json_response(response.dict())


@router.post(
    "/conversation/{conversation_id}/context",
    tags=[TAG],
    summary="Send additional context for conversation.",
    response_model=ApiResponse,
)
async def post_context(
        conversation_id: str = Path(..., description="The ID of the conversation."),
        payload: ChatContextRequest = Body(..., description="User context")
):
    context_repository = ConversationUserContextRepositoryLocal()
    response = chat_context_service.execute(
        conversation_id=conversation_id,
        request=payload,
        repository=context_repository)
    return routing_utils.to_json_response(response.dict())


@router.post(
    "/conversation_stream/{conversation_id}",
    summary="Send message to the base copilot and get the response as a stream.",
    response_description=STREAM_RESPONSE_DESCRIPTION,
    tags=[TAG]
)
async def handle_conversation_streaming(
        email: Optional[str] = Header(default=None),
        conversation_id: str = Path(..., description="The ID of the conversation."),
        payload: ConversationInput = Body(...,
                                          description="Input and parameters for the conversation."),
        user_id: str = Depends(validate_api_key_use_case.execute)
):
    request = ChatRequest(
        chat_id=conversation_id,
        message=payload.inputs,
        response_message_id=payload.response_message_id,
        email=user_id or email
    )

    history_repository = ConversationHistoryRepositoryLocal()
    logs_repository = ConversationLogsRepositoryLocal()

    headers = {
        'X-Content-Type-Options': 'nosniff',
        'Connection': 'keep-alive',
    }
    return StreamingResponse(
        chat_streaming_service.execute(
            request,
            document_store.get_document_store(),
            history_repository,
            logs_repository,
        ),
        headers=headers,
        media_type="text/event-stream"
    )


@router.get(
    "/conversation/{conversation_id}/history",
    summary="Retrieve conversation history.",
    tags=[TAG]
)
async def handle_get_conversation_history(
        conversation_id: str = Path(..., description="The ID of the conversation."),
):
    request = ChatHistoryRequest(
        chat_id=conversation_id,
    )

    history_repository = ConversationHistoryRepositoryLocal()

    response: ChatHistoryResponse = await chat_history_service.execute(
        request,
        history_repository,
    )
    return response



import uuid

from opencopilot.domain.chat import on_user_message_use_case
from opencopilot.domain.chat.entities import UserMessageInput
from opencopilot.repository.conversation_history_repository import ConversationHistoryRepositoryLocal
from opencopilot.repository.conversation_logs_repository import ConversationLogsRepositoryLocal
from opencopilot.repository.documents.document_store import DocumentStore
from opencopilot.service.chat.entities import ChatRequest
from opencopilot.service.chat.entities import ChatResponse
from opencopilot.service.utils import get_uuid


async def execute(
        request: ChatRequest,
        document_store: DocumentStore,
        history_repository: ConversationHistoryRepositoryLocal,
        logs_repository: ConversationLogsRepositoryLocal,
) -> ChatResponse:
    chat_id = get_uuid(request.chat_id, "chat_id")
    domain_response = await on_user_message_use_case.execute(
        UserMessageInput(
            chat_id=chat_id,
            message=request.message,
            response_message_id=request.response_message_id or str(uuid.uuid4()),
            email=request.email
        ),
        document_store,
        history_repository,
        logs_repository=logs_repository,
    )
    return ChatResponse(
        response="OK",
        chat_id=str(chat_id),
        message=domain_response.content,
        sources=domain_response.sources,
    )



import time
from typing import List
from typing import Optional
from urllib.parse import urljoin

from bs4 import BeautifulSoup
from langchain.schema import Document
from playwright.sync_api import sync_playwright


def execute(urls: List[str]) -> List[Document]:
    main_urls = _get_main_urls(urls)
    main_urls = list(set(main_urls))
    docs = _scrape_url_contents(main_urls)
    return docs


def _get_main_urls(urls: List[str]) -> List[str]:
    links: List[str] = []
    for url in urls:
        soup = _scrape_html_with_playwright(url)
        if soup:
            new_links = _find_links(soup, url)
            links += new_links
    return links


def _scrape_html_with_playwright(url) -> Optional[BeautifulSoup]:
    soup = None
    try:
        with sync_playwright() as p:
            browser = p.chromium.launch(headless=True)
            try:
                page = browser.new_page()
                page.goto(url)
                page_source = page.content()
                soup = BeautifulSoup(page_source, features="lxml")
            except:
                pass
            browser.close()
    except Exception as e:
        print(e)
    return soup


def _find_links(
        soup: BeautifulSoup,
        base_url: str,
) -> List[str]:
    links: List[str] = []
    tags = soup.find_all("a")
    for tag in tags:
        if href := tag.attrs.get("href"):
            if "#" not in href and not href.startswith("http") and not href.startswith(".") and href.endswith("/"):
                links.append(urljoin(base_url, href))
    return links


def _scrape_url_contents(all_urls: List[str]) -> List[Document]:
    docs: List[Document] = []
    count = 0
    for url in all_urls:
        soup = _scrape_html_with_playwright(url)
        tags = soup.find_all("div", class_="md-content")
        if len(tags):
            docs.append(Document(
                page_content=tags[0].text,
                metadata={
                    "source": url,
                }
            ))
        else:
            print(f"Failed for url: {url}")
        time.sleep(1.5)
        count += 1
        if count % 10 == 0:
            print(f"{count}/{len(all_urls)}")
    return docs



import json
import uuid

from opencopilot.domain.chat import on_user_message_streaming_use_case
from opencopilot.domain.chat.entities import UserMessageInput
from opencopilot.repository.conversation_history_repository import ConversationHistoryRepositoryLocal
from opencopilot.repository.conversation_logs_repository import ConversationLogsRepositoryLocal
from opencopilot.repository.documents.document_store import DocumentStore
from opencopilot.service.chat.entities import ChatRequest
from opencopilot.service.utils import get_uuid


async def execute(
        request: ChatRequest,
        document_store: DocumentStore,
        history_repository: ConversationHistoryRepositoryLocal,
        logs_repository: ConversationLogsRepositoryLocal,
) -> str:
    chat_id = get_uuid(request.chat_id, "chat_id")
    async for chunk in on_user_message_streaming_use_case.execute(
            UserMessageInput(
                chat_id=chat_id,
                message=request.message,
                response_message_id=request.response_message_id or str(uuid.uuid4()),
                email=request.email
            ),
            document_store,
            history_repository,
            logs_repository,
    ):
        data = chunk.to_dict()
        yield f"{json.dumps(data)}\n"



from fastapi import FastAPI
from fastapi.openapi.utils import get_openapi
from pydantic import BaseModel
from starlette.middleware.cors import CORSMiddleware

from opencopilot import settings
from opencopilot.routers import main_router, routing_utils
from opencopilot.service.exception_handlers.exception_handlers import custom_exception_handler
from opencopilot.service.middleware.main_middleware import MainMiddleware
from opencopilot.service.middleware.request_enrichment_middleware import RequestEnrichmentMiddleware

app = FastAPI()

app.include_router(
    main_router.router,
    prefix="/v0"
)

API_TITLE = "API"
API_DESCRIPTION = "API"
API_VERSION = "0.1"


class ApiInfo(BaseModel):
    title: str
    description: str
    version: str

    class Config:
        schema_extra = {
            "example": {
                "title": API_TITLE,
                "description": API_DESCRIPTION,
                "version": API_VERSION
            }
        }


def custom_openapi():
    if app.openapi_schema:
        return app.openapi_schema

    openapi_schema = get_openapi(
        title="API",
        version="1.0.0",
        description="API version 1.0.0",
        routes=app.routes,
        servers=_get_servers()
    )
    openapi_schema["info"]["contact"] = {
        "name": "",
        "email": ""
    }
    openapi_schema["info"]["x-logo"] = {
        "url": ""
    }
    openapi_schema["x-readme"] = {
        "samples-languages": [
            "curl",
            "node",
            "javascript",
            "python"
        ]
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


def _get_servers():
    servers = []
    if settings.get().is_production():
        pass
    else:
        base_url = settings.get().API_BASE_URL
        if base_url.endswith("/"):
            base_url = base_url[:-1]
        servers.append({"url": f"{base_url}:{settings.get().API_PORT}"})
    return servers


app.openapi = custom_openapi

# order of middleware matters! first middleware called is the last one added
app.add_middleware(
    CORSMiddleware,
    allow_origins=settings.get().ALLOWED_ORIGINS,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.add_middleware(MainMiddleware)
app.add_middleware(RequestEnrichmentMiddleware)

# exception handlers run AFTER the middlewares!
# Handles API error responses
app.add_exception_handler(Exception, custom_exception_handler)


# Overrides FastAPI error responses, eg: authorization, not found
# app.add_exception_handler(StarletteHTTPException, custom_http_exception_handler)
# Overrides default Pydantic request validation errors
# app.add_exception_handler(RequestValidationError, validation_exception_handler)


def get_api_info() -> ApiInfo:
    return ApiInfo(
        title=API_VERSION,
        description=API_DESCRIPTION,
        version=API_VERSION
    )


@app.get(
    "/",
    summary="Returns API information",
    description="Returns API information",
    response_description="API information with title, description and version.",
    response_model=ApiInfo,
    include_in_schema=not settings.get().is_production())
def root():
    return routing_utils.to_json_response(get_api_info().dict())



from pydantic import BaseModel
from pydantic import Field

from opencopilot.service.entities import ApiResponse


class TokenRequest(BaseModel):
    client_id: str = Field(
        description="Client id"
    )
    client_secret: str = Field(
        description="Client secret"
    )
    user_id: str = Field(
        description="User id"
    )


class TokenResponse(ApiResponse):
    token: str = Field(
        description="JWT token"
    )

    class Config:
        schema_extra = {
            "example": {
                "response": "OK",
                "token": "e91042aa-d53a-41eb-8884-67aa4947982d",
            }
        }



import asyncio
import json
from datetime import datetime
from typing import AsyncGenerator
from typing import List

from langchain.schema import Document

from opencopilot import settings
from opencopilot.logger import api_logger
from opencopilot.domain.chat import validate_urls_use_case
from opencopilot.domain.chat.entities import LoadingMessage
from opencopilot.domain.chat.entities import StreamingChunk
from opencopilot.domain.chat.entities import UserMessageInput
from opencopilot.domain.chat.results import get_gpt_result_use_case
from opencopilot.domain.chat.utils import get_system_message
from opencopilot.repository.conversation_history_repository import ConversationHistoryRepositoryLocal
from opencopilot.repository.conversation_logs_repository import ConversationLogsRepositoryLocal
from opencopilot.repository.documents.document_store import DocumentStore
from opencopilot.utils.callbacks.callback_handler import CustomAsyncIteratorCallbackHandler

logger = api_logger.get()


async def execute(
        domain_input: UserMessageInput,
        document_store: DocumentStore,
        history_repository: ConversationHistoryRepositoryLocal,
        logs_repository: ConversationLogsRepositoryLocal,
) -> AsyncGenerator[StreamingChunk, None]:
    system_message = get_system_message()

    context = _get_context(
        domain_input,
        system_message,
        document_store
    )
    message_timestamp = datetime.now().timestamp()

    callback = CustomAsyncIteratorCallbackHandler()

    task = asyncio.create_task(
        get_gpt_result_use_case.execute(
            domain_input,
            system_message,
            context,
            logs_repository=logs_repository,
            history_repository=history_repository,
            callback=callback,
        )
    )

    result = ""
    try:
        async for callback_result in callback.aiter():
            parsed = json.loads(callback_result)
            if token := parsed.get("token"):
                yield StreamingChunk(
                    chat_id=domain_input.chat_id,
                    text=token,
                    sources=[],
                )
            if loading_message := parsed.get("loading_message"):
                yield StreamingChunk(
                    chat_id=domain_input.chat_id,
                    text="",
                    sources=[],
                    loading_message=LoadingMessage.from_dict(loading_message),
                )
        await task
        result = task.result()
    except Exception as exc:
        logger.error(f"Stream error: {exc}")
        yield StreamingChunk(
            chat_id=domain_input.chat_id,
            text="",
            sources=[],
            error=f"OpenAI error: {type(exc).__name__}",
        )
    finally:
        validate_urls_use_case.execute(result, domain_input.chat_id)

        response_timestamp = datetime.now().timestamp()

        history_repository.save_history(
            domain_input.message,
            result,
            message_timestamp,
            response_timestamp,
            domain_input.chat_id,
            domain_input.response_message_id,
        )


def _get_context(
        domain_input: UserMessageInput,
        system_message: str,
        document_store: DocumentStore
) -> List[Document]:
    # TODO: handle context length and all the edge cases somehow a bit better
    context = []
    if "{context}" in system_message:
        context = []
        context.extend(
            document_store.find(
                domain_input.message,
                k=settings.get().MAX_CONTEXT_DOCUMENTS_COUNT - len(context)
            )
        )
    return context



from dataclasses import dataclass
from typing import List

from dataclasses_json import dataclass_json


@dataclass_json
@dataclass(frozen=True)
class RetrievalExample:
    # User query
    query: str
    # Relevant documents
    documents: List[str]


@dataclass_json
@dataclass(frozen=True)
class RetrievalDataset:
    examples: List[RetrievalExample]


@dataclass_json
@dataclass(frozen=True)
class RetrievalResult:
    # Returned documents
    documents: List[str]


@dataclass_json
@dataclass(frozen=True)
class RetrievalSingleEvaluation:
    example: RetrievalExample
    result: RetrievalResult
    true_positives: List[str]
    false_positives: List[str]
    false_negatives: List[str]
    precision: float
    recall: float


@dataclass_json
@dataclass(frozen=True)
class RetrievalSummaryEvaluation:
    single_evaluations: List[RetrievalSingleEvaluation]
    average_precision: float
    average_recall: float


@dataclass_json
@dataclass(frozen=True)
class EndToEndExample:
    query: str
    answer: str


@dataclass_json
@dataclass(frozen=True)
class EndToEndResult:
    answer: str
    documents: List[str]


@dataclass_json
@dataclass(frozen=True)
class EndToEndDataset:
    examples: List[EndToEndExample]


@dataclass_json
@dataclass(frozen=True)
class EndToEndSingleEvaluation:
    evaluation: str


@dataclass_json
@dataclass(frozen=True)
class EndToEndSummaryEvaluation:
    single_evaluations: List[EndToEndSingleEvaluation]
    evaluations_count: int
    evaluations_score: float



import jwt
import time

from opencopilot import settings
from opencopilot.service import error_responses


def execute(client_id: str, client_secret: str, user_id: str) -> str:
    if client_id != settings.get().JWT_CLIENT_ID or client_secret != settings.get().JWT_CLIENT_SECRET:
        raise error_responses.InvalidCredentialsAPIError

    payload = {
            'iat': time.time() - 100,
            'exp': time.time() + settings.get().JWT_TOKEN_EXPIRATION_SECONDS,
            'sub': user_id
        }
    return jwt.encode(payload, settings.get().JWT_CLIENT_SECRET, algorithm='HS256')



from dataclasses import dataclass


@dataclass(frozen=True)
class EvaluationInput:
    query: str
    answer: str
    expected_answer: str


@dataclass(frozen=True)
class EvaluationOutput:
    evaluation: str



from typing import Optional

from opencopilot.domain.debug import message_debug_use_case
from opencopilot.domain.debug.entities import TextWithTokens
from opencopilot.repository.conversation_history_repository import ConversationHistoryRepositoryLocal
from opencopilot.repository.conversation_logs_repository import ConversationLogsRepositoryLocal
from opencopilot.service import utils
from opencopilot.service.debug.entities import GetMessageDebugResponse
from opencopilot.service.debug.entities import ValueWithTokens


def execute(
        conversation_id: str,
        message_id: str,
        history_repository: ConversationHistoryRepositoryLocal,
        logs_repository: ConversationLogsRepositoryLocal,
) -> GetMessageDebugResponse:
    domain_response = message_debug_use_case.execute(
        utils.get_uuid(conversation_id, "conversation_id"),
        message_id,
        history_repository,
        logs_repository,
    )
    return GetMessageDebugResponse(
        response="OK",
        prompt_template=_convert(domain_response.prompt_template),
        data_sources=domain_response.data_sources,
        user_question=_convert(domain_response.user_question),
        context=_convert(domain_response.context),
        chat_history=_convert(domain_response.chat_history),
        full_prompt=_convert(domain_response.full_prompt),
        llm_response=_convert(domain_response.llm_response),
    )


def _convert(value: Optional[TextWithTokens]) -> Optional[ValueWithTokens]:
    if not value:
        return None
    return ValueWithTokens(
        value=value.text,
        token_count=value.token_count,
    )



import re
import string
from typing import List
from uuid import UUID

import requests

from opencopilot.logger import api_logger
from opencopilot.repository.slack import slack_messenger

URL_REGEX = r"""((?:(?:https|ftp|http)?:(?:/{1,3}|[a-z0-9%])|[a-z0-9.\-]+[.](?:com|org|uk)/)(?:[^\s()<>{}\[\]]+|\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\))+(?:\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\)|[^\s`!()\[\]{};:'".,<>?«»“”‘’])|(?:(?<!@)[a-z0-9]+(?:[.\-][a-z0-9]+)*[.](?:com|uk|ac)\b/?(?!@)))"""

logger = api_logger.get()


def execute(text: str, chat_id: UUID) -> None:
    urls: List[str] = _find_urls(text)
    for url in urls:
        _validate_url(url, chat_id)


def _find_urls(text: str) -> List[str]:
    urls = re.findall(URL_REGEX, text)
    return [''.join(x for x in url if x in string.printable) for url in urls]


def _validate_url(url: str, chat_id: UUID) -> None:
    try:
        response = requests.get(url)
        if response.status_code in [404, 410]:
            logger.warning(
                f"Error querying url {url}, conversation_id: {str(chat_id)}")
            slack_messenger.post_error(
                f"Error querying url {url}, conversation_id: {str(chat_id)}",
                f"status_code {response.status_code}",
            )
    except Exception as e:
        logger.warning(f"Error querying url {url}, conversation_id: {str(chat_id)}")
        slack_messenger.post_error(
            f"Error querying url {url}, conversation_id: {str(chat_id)}",
            str(e),
        )



import json
import os
from dataclasses import dataclass
from typing import Literal
from typing import Optional
from typing import Union

from omegaconf import DictConfig
from omegaconf import ListConfig
from omegaconf import OmegaConf


@dataclass(frozen=False)
class Settings:
    COPILOT_NAME: str

    API_PORT: int
    API_BASE_URL: str
    ENVIRONMENT: str
    ALLOWED_ORIGINS: str
    APPLICATION_NAME: str
    LOG_FILE_PATH: str

    WEAVIATE_URL: str
    WEAVIATE_READ_TIMEOUT: int

    MODEL: Literal["gpt-3.5-turbo-16k", "gpt-4"]

    OPENAI_API_KEY: str

    MAX_DOCUMENT_SIZE_MB: int

    SLACK_WEBHOOK: str

    AUTH_TYPE: Optional[str]
    API_KEY: str

    JWT_CLIENT_ID: str
    JWT_CLIENT_SECRET: str
    JWT_TOKEN_EXPIRATION_SECONDS: int

    HELICONE_API_KEY: str
    HELICONE_RATE_LIMIT_POLICY: str

    CONVERSATIONS_DIR: str = "conversations"
    # Configure based on model?
    PROMPT_HISTORY_INCLUDED_COUNT: int = 4
    MAX_CONTEXT_DOCUMENTS_COUNT: int = 4
    MAX_TOKEN_COUNT: int = 2048

    PROMPT_FILE: Optional[str] = None

    DEFAULT_PROMPT: str = """Your are a Basketball Copilot. You are an expert on basketball.

    =========
    {context}
    =========

    {history}
    User: {question}
    Copilot answer in Markdown:"""

    COPILOTS_ROOT_PATH = "copilots"

    HELICONE_BASE_URL = "https://oai.hconeai.com/v1"

    DATA_DIR: str = ""

    def __post_init__(self):
        os.environ["OPENAI_API_KEY"] = self.OPENAI_API_KEY

        if self.AUTH_TYPE is not None and (self.AUTH_TYPE == "none" or self.AUTH_TYPE == "None" or self.AUTH_TYPE.strip() == ""):
            self.AUTH_TYPE = None

        self.COPILOT_DIRECTORY = f"copilots/{self.COPILOT_NAME}"
        self.copilot_config: Union[DictConfig, ListConfig, None] = None

        self.PROMPT_QUESTION_KEY = self._get_prompt_key("question_key") or "User"
        self.PROMPT_ANSWER_KEY = self._get_prompt_key("response_key") or "Copilot"

    def is_production(self):
        return self.ENVIRONMENT == "production"

    def get_max_token_count(self) -> int:
        if self.MODEL == "gpt-3.5-turbo-16k":
            return 16384
        if self.MODEL == "gpt-4":
            return 8192
        return 2048

    def _get_prompt_key(self, key: str) -> Optional[str]:
        try:
            with open(f"{self.COPILOT_DIRECTORY}/prompts/prompt_configuration.json", "r") as f:
                prompt_configuration = json.load(f)
            return prompt_configuration.get(key) or None
        except:
            pass
        return None


def init_data_dir(data_dir: str) -> None:
    settings = get()
    if settings:
        settings.DATA_DIR = data_dir if os.path.exists(data_dir) else None


def init_custom_loaders(config_file: str) -> None:
    settings = get()
    if settings:
        try:
            settings.copilot_config = OmegaConf.load(config_file)
        except:
            pass


def init_prompt_file_location(file_path: str) -> None:
    settings = get()
    if settings:
        if os.path.isfile(file_path):
            settings.PROMPT_FILE = file_path


_settings: Optional[Settings] = None


def get() -> Optional[Settings]:
    global _settings
    return _settings


def set(new_settings: Settings):
    global _settings
    _settings = new_settings



from pydantic import BaseModel
from pydantic import Field


class ApiResponse(BaseModel):
    response: str = Field(
        description="Response status, either OK or NOK."
    )

    class Config:
        schema_extra = {
            "example": {
                "response": "OK"
            }
        }



from opencopilot.domain.chat.entities import ChatFeedbackInput
from opencopilot.domain.chat.entities import ChatFeedbackOutput
from opencopilot.repository.conversation_history_repository import ConversationHistoryRepositoryLocal


def execute(
        data_input: ChatFeedbackInput,
        repository: ConversationHistoryRepositoryLocal
) -> ChatFeedbackOutput:
    repository.add_feedback(data_input)
    return ChatFeedbackOutput(
        response="OK"
    )



from typing import List

from langchain.schema import Document


def execute(documents: List[Document]) -> str:
    formatted_documents: List[str] = []
    for d in documents:
        document_strings: List[str] = []
        if title := d.metadata.get("title"):
            document_strings.append(f"Title: {title}")
        if source := d.metadata.get("source"):
            document_strings.append(f"Source: {source}")
        document_strings.append("Content:\n" + d.page_content)
        formatted_documents.append("\n".join(document_strings))
    return "\n\n".join(formatted_documents)



from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class TextWithTokens:
    text: Optional[str]
    token_count: Optional[int]


@dataclass(frozen=True)
class MessageDebugResult:
    prompt_template: Optional[TextWithTokens]
    data_sources: Optional[str]
    user_question: Optional[TextWithTokens]
    context: Optional[TextWithTokens]
    chat_history: Optional[TextWithTokens]
    full_prompt: Optional[TextWithTokens]
    # Need more data here
    llm_response: Optional[TextWithTokens]



from typing import Dict, Optional

import openai
from langchain.chat_models import ChatOpenAI

from opencopilot import settings
from opencopilot.utils.callbacks.callback_handler import CustomAsyncIteratorCallbackHandler


def execute(
        email: str = None,
        callback: CustomAsyncIteratorCallbackHandler = None,
) -> ChatOpenAI:
    if settings.get().HELICONE_API_KEY:
        openai.api_base = settings.get().HELICONE_BASE_URL
    llm = ChatOpenAI(
        temperature=0.0,
        model_name=settings.get().MODEL,
        streaming=callback is not None,
        callbacks=[callback] if callback is not None else None,
        headers=_get_headers(email)
    )
    return llm


def _get_headers(email: str = None) -> Optional[Dict]:
    if settings.get().HELICONE_API_KEY:
        headers = {
            "Helicone-Auth": "Bearer " + settings.get().HELICONE_API_KEY,
            "Helicone-User-Id": email or "",
        }
        if email and settings.get().HELICONE_RATE_LIMIT_POLICY:
            headers["Helicone-RateLimit-Policy"] = settings.get().HELICONE_RATE_LIMIT_POLICY
        return headers
    return None



from opencopilot.domain.chat import chat_feedback_use_case
from opencopilot.domain.chat.entities import ChatFeedbackInput
from opencopilot.repository.conversation_history_repository import \
    ConversationHistoryRepositoryLocal
from opencopilot.service.chat.entities import ChatFeedbackRequest
from opencopilot.service.entities import ApiResponse
from opencopilot.service.utils import get_uuid


def execute(
        conversation_id: str,
        request: ChatFeedbackRequest,
        repository: ConversationHistoryRepositoryLocal
) -> ApiResponse:
    conversation_id = get_uuid(conversation_id, "conversation_id")
    response = chat_feedback_use_case.execute(
        data_input=ChatFeedbackInput(
            conversation_id=conversation_id,
            correctness=request.correctness,
            helpfulness=request.helpfulness,
            easy_to_understand=request.easy_to_understand,
            free_form_feedback=request.free_form_feedback
        ),
        repository=repository
    )
    return ApiResponse(
        response=response.response
    )



import time

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

from opencopilot.logger import api_logger
from opencopilot.service.error_responses import APIErrorResponse
from opencopilot.service.middleware import util
from opencopilot.service.middleware.entitites import RequestStateKey
from opencopilot.utils.http_headers import add_response_headers

logger = api_logger.get()


class MainMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp) -> None:
        super().__init__(app)

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        request_id = util.get_state(request, RequestStateKey.REQUEST_ID)
        ip_address = util.get_state(request, RequestStateKey.IP_ADDRESS)
        country = util.get_state(request, RequestStateKey.COUNTRY)
        try:
            logger.info(
                f"REQUEST STARTED "
                f"request_id={request_id} "
                f"request_path={request.url.path} "
                f"ip={ip_address} "
                f"country={country} "
            )
            before = time.time()
            response: Response = await call_next(request)

            duration = time.time() - before
            process_time = (time.time() - before) * 1000
            formatted_process_time = '{0:.2f}'.format(process_time)
            if response.status_code != 404:
                logger.info(
                    f"REQUEST COMPLETED "
                    f"request_id={request_id} "
                    f"request_path={request.url.path} "
                    f"completed_in={formatted_process_time}ms "
                    f"status_code={response.status_code}"
                )
            return await _get_response_with_headers(response, duration)
        except Exception as error:
            if isinstance(error, APIErrorResponse):
                is_exc_info = error.to_status_code() == 500
                logger.error(
                    f"Error while handling request. request_id={request_id} "
                    f"request_path={request.url.path}"
                    f"status code={error.to_status_code()}"
                    f"code={error.to_code()}"
                    f"message={error.to_message()}",
                    exc_info=is_exc_info
                )
            else:
                logger.error(
                    f"Error while handling request. request_id={request_id} "
                    f"request_path={request.url.path}",
                    exc_info=True)
            raise error


async def _get_response_with_headers(response, duration):
    response = await add_response_headers(response, duration)
    return response



from fastapi import APIRouter
from fastapi import Body
from opencopilot.logger import api_logger
from opencopilot.service.authorization import token_service
from opencopilot.service.authorization.entities import TokenRequest
from opencopilot.service.authorization.entities import TokenResponse

TAG = "Authorization"
router = APIRouter()
router.openapi_tags = [TAG]
router.title = "Authorization router"

logger = api_logger.get()


@router.post(
    "/token",
    tags=[TAG],
    summary="Generate token.",
    response_model=TokenResponse,
)
async def evaluate(
        request: TokenRequest = Body(
            ...,
            description="Token generation input")
):
    return token_service.execute(request)



from typing import List
from typing import Sequence

from langchain.callbacks.manager import Callbacks
from langchain.chat_models import ChatOpenAI
from langchain.evaluation import QAEvalChain

from opencopilot.domain.evaluation.entities import EvaluationOutput
from opencopilot.domain.evaluation.entities import EvaluationInput
from opencopilot.eval import endtoend


async def execute(domain_input: EvaluationInput) -> EvaluationOutput:
    llm = ChatOpenAI(temperature=0, model_name="gpt-4")

    eval_chain = AsyncQAEvalChain.from_llm(llm=llm, prompt=endtoend.PROMPT)
    eval_example = {"query": domain_input.query, "answer": domain_input.answer}
    eval_prediction = {"result": domain_input.expected_answer}

    graded_output = (await eval_chain.aevaluate([eval_example], [eval_prediction]))[0]

    return EvaluationOutput(evaluation=graded_output["text"])


class AsyncQAEvalChain(QAEvalChain):
    async def aevaluate(
            self,
            examples: Sequence[dict],
            predictions: Sequence[dict],
            question_key: str = "query",
            answer_key: str = "answer",
            prediction_key: str = "result",
            *,
            callbacks: Callbacks = None,
    ) -> List[dict]:
        """Evaluate question answering examples and predictions."""
        inputs = [
            {
                "query": example[question_key],
                "answer": example[answer_key],
                "result": predictions[i][prediction_key],
            }
            for i, example in enumerate(examples)
        ]

        return await self.aapply(inputs, callbacks=callbacks)



"""
Search evaluation script.

Currently, the script expects that Weaviate vector store is running at http://0.0.0.0:8080/ and is
already filled with data.

If nothing is retrieved and nothing should be retrieved, then the result is perfect -
100% precision and 100% recall. But mostly, something should be retrieved. In that case:
If document retrieval returns nothing, the precision is 100%, and recall is 0%.
If document retrieval return everything, the precision is near 0%, and the recall is 100%.

Usage: see python eval_retrieval.py --help

"""
import argparse
import json
from typing import Callable
from typing import List


import matplotlib.pyplot as plt
from pathlib import Path
from langchain.schema import Document
from opencopilot.repository.documents import document_store

from tqdm import tqdm

from opencopilot import settings
from opencopilot.eval.retrieval import evaluate_retrieval_dataset
from opencopilot.eval.entities import (
    RetrievalDataset,
    RetrievalResult,
    RetrievalSummaryEvaluation,
)


def parse_uri(s: str):
    """Turn into a (relatively) canonical URI."""
    s = s.strip()

    if "://" in s:
        return s[s.index("://") + 3 :]

    return s


def dataset_from_file(dataset_path: Path, limit: int=None) -> RetrievalDataset:
    with open(dataset_path) as f:
        examples_dict = json.load(f)
        if limit:
            examples_dict["examples"] = examples_dict["examples"][:limit]

        # Fix URLs - todo this is a bit hacky but best place to canonicalize URLs
        for example in examples_dict["examples"]:
            example["documents"] = [parse_uri(d) for d in example["documents"]]

    return RetrievalDataset.from_dict(examples_dict)


def print_metrics(metrics: RetrievalSummaryEvaluation):
    print("  Metrics")
    print(f"    average precision:", metrics.average_precision)
    print(f"       average recall:", metrics.average_recall)


def evaluate_retriever(
    search_function: Callable, dataset: RetrievalDataset, **search_kwargs
) -> RetrievalSummaryEvaluation:
    """Evaluate a retriever function on a dataset."""
    

    predictions = []
    for example in tqdm(dataset.examples):
        predicted_raw: List[Document] = search_function(example.query, **search_kwargs)
        retrieved_urls = [parse_uri(p.metadata["source"]) for p in predicted_raw]
        predictions.append(RetrievalResult(documents=retrieved_urls))

    return evaluate_retrieval_dataset(dataset, predictions)


def _draw_curve(dataset_path: str, limit=None):
    final_metrics = []
    for k in range(0, 11):
        document_retriever = document_store.get_document_store()
        search_type = f"similarity_search"
        metrics = evaluate_retriever(
            document_retriever.find,
            dataset = dataset_from_file(dataset_path, limit=limit),
            search_type=search_type,
            k=k + 1,
        )
        final_metrics.append(metrics)
        print_metrics(metrics)

    _plot_curve(final_metrics)


def _plot_curve(final_metrics: List[RetrievalSummaryEvaluation]):
    x = [item for item in range(len(final_metrics))]
    y1 = [item.average_precision for item in final_metrics]
    y2 = [item.average_recall for item in final_metrics]
    plt.plot(x, y1, label="average precision")
    plt.plot(x, y2, label="average recall")
    plt.xlabel("k")
    plt.ylabel("Average Precision & Recall")
    plt.title("Average Precision & Recall")
    plt.legend(loc="upper left")
    plt.show()


def _log_wandb(summary_evaluation: RetrievalSummaryEvaluation):
    import wandb

    wandb.init(
        # Set the project where this run will be logged
        project="opencopilot-retrieval",
        # Track hyperparameters and run metadata
        config={},
    )
    wandb.log(
        {
            "average_precision": summary_evaluation.average_precision,
            "average_recall": summary_evaluation.average_recall,
        }
    )
    with open(wandb.run.dir + "/evaluations.json", "w") as f:
        json.dump(summary_evaluation.to_dict(), f, indent=4)

    wandb.finish()


def main(dataset_path: str, draw_curve: bool, output_path: str = None, limit: int = None):
    # TODO: document_store.init_document_store()
    if draw_curve:
        _draw_curve(dataset_path, limit=limit)
    else:
        document_retriever = document_store.get_document_store()
        metrics = evaluate_retriever(
            document_retriever.find,
            dataset = dataset_from_file(dataset_path, limit=limit),
            search_type="similarity_search",
            k=4,
        )
        print_metrics(metrics)
        if output_path:
            with open(output_path, "w") as f:
                json.dump(metrics.to_dict(), f, indent=4)
        return metrics


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--dataset_path", type=str, default=f"../copilots/{settings.get().COPILOT_NAME}/eval_data/retrieval_human.json"
    )
    parser.add_argument(
        "-n", "--num_examples", type=int, default=None, help="Limit how many examples are evaluated from dataset. Default: all examples used."
    )
    parser.add_argument(
        "--draw_curve", action="store_true", help="Draw precision/recall curve. Disables --wandb and --output."
    )
    parser.add_argument("--wandb", action="store_true", help="Store results in WANDB")
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=None,
        help="Output file for evaluation results in JSON format.",
    )
    args = parser.parse_args()

    metrics = main(args.dataset_path, args.draw_curve, output_path=args.output, limit=args.num_examples)

    if args.wandb and not args.draw_curve:
        print("Logging to WANDB:")
        _log_wandb(metrics)



from opencopilot.logger import api_logger
from opencopilot.repository.conversation_history_repository import ConversationHistoryRepositoryLocal
from opencopilot.service.chat.entities import ChatHistoryItem

logger = api_logger.get()


async def execute(
        chat_id: str,
        history_repository: ConversationHistoryRepositoryLocal,
) -> [ChatHistoryItem]:
    response = history_repository.get_history(
        chat_id
    )
    return_value = []
    timer = 0
    for message in response:
        prompt_timestamp = message['prompt_timestamp'] if 'prompt_timestamp' in message else timer
        response_timestamp = message['response_timestamp'] if 'response_timestamp' in message else timer + 1
        timer = max(timer, prompt_timestamp, response_timestamp) + 2
        return_value = return_value + [ChatHistoryItem(content=message['prompt'], timestamp=prompt_timestamp), ChatHistoryItem(content=message['response'], timestamp=response_timestamp)]

    return return_value



import json
from typing import Dict
from typing import List
from typing import Optional
from uuid import UUID

from opencopilot import settings
from opencopilot.logger import api_logger
from opencopilot.domain.chat.entities import ChatFeedbackInput

DEFAULT_CONVERSATIONS_DIR = "conversations"

logger = api_logger.get()


class ConversationHistoryRepositoryLocal:

    def __init__(
            self,
            conversations_dir: str = DEFAULT_CONVERSATIONS_DIR,
            question_key: str = "",
            response_key: str = "",
    ):
        self.conversations_dir = conversations_dir
        self.question_key = question_key or settings.get().PROMPT_QUESTION_KEY
        self.response_key = response_key or settings.get().PROMPT_ANSWER_KEY

    def get_prompt_history(self, chat_id: UUID, count: Optional[int]) -> str:
        try:
            with open(self._get_file_path(chat_id), "r") as f:
                history = json.load(f)
            if not count or len(history) <= count:
                return self._to_string(history)
            return self._to_string(history[count * -1:])
        except:
            logger.debug(f"Cannot load conversation history, id: {str(chat_id)}")
        return ""

    def get_history(self, chat_id: UUID) -> List[Dict]:
        history = []
        try:
            with open(self._get_file_path(chat_id), "r") as f:
                history = json.load(f)
        except:
            pass
        return history

    def _to_string(self, history: List[Dict]) -> str:
        formatted: str = ""
        for i in history:
            formatted += f"{self.question_key}: {i.get('prompt', '')}\n"
            formatted += f"{self.response_key}: {i.get('response', '')}\n"
        return formatted

    def save_history(
            self,
            message: str,
            result: str,
            prompt_timestamp: float,
            response_timestamp: float,
            chat_id: UUID,
            response_message_id: str,
    ) -> None:
        history = self.get_history(chat_id)
        history.append({
            "prompt": message,
            "response": result.strip(),
            "prompt_timestamp": prompt_timestamp,
            "response_timestamp": response_timestamp,
            "response_message_id": response_message_id,
        })
        self._write_file(chat_id, history)

    def add_feedback(
            self,
            chat_feedback: ChatFeedbackInput
    ) -> None:
        history = self.get_history(chat_feedback.conversation_id)
        history[-1]["user_feedback"] = {
            "correctness": chat_feedback.correctness,
            "helpfulness": chat_feedback.helpfulness,
            "easy_to_understand": chat_feedback.easy_to_understand,
            "free_form_feedback": chat_feedback.free_form_feedback,
        }
        self._write_file(chat_feedback.conversation_id, history)

    def _get_file_path(self, chat_id: UUID):
        return f"{self.conversations_dir}/{str(chat_id)}.json"

    def _write_file(self, chat_id: UUID, data):
        try:
            with open(self._get_file_path(chat_id), "w") as f:
                f.write(json.dumps(data, indent=4))
        except Exception as e:
            logger.warning(f"Failed to save history for chat {str(chat_id)}")



import os

from setuptools import setup


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


extra_files = package_files('opencopilot')

setup(
    name='opencopilot-ai',
    version='0.3.0',
    packages=["opencopilot"],
    package_data={'': extra_files},
    license="MIT",
    description="OpenCopilot Backend",
    author="OpenCopilot",
    author_email="kaspar@nftport.xyz",
    url="https://github.com/opencopilotdev/opencopilot",
    py_modules=['opencopilot'],
    install_requires=[
        'fastapi',
        'python-dotenv',
        'python-json-logger',
        'uvicorn',
        'pandas',
        'pexpect',
        'langchain>=0.0.217',
        'passlib',
        'pyjwt[crypto]',
        'Jinja2',
        'tiktoken',
        'text-generation',
        'weaviate-client',
        'pypdf',
        'unstructured',
        'pdf2image',
        'beautifulsoup4',
        'openai',
        'unstructured',
        'dataclasses-json',
        'pytesseract',
        'aiohttp',
        'xxhash',
        'omegaconf',
    ],
)



from typing import Optional

from pydantic import BaseModel
from pydantic import Field

from opencopilot.service.entities import ApiResponse


class ValueWithTokens(BaseModel):
    value: str = Field()
    token_count: Optional[int] = Field(
        description="If debug mode is enabled will include token counts."
    )


class GetMessageDebugResponse(ApiResponse):
    prompt_template: Optional[ValueWithTokens] = Field(
        description="Prompt template used."
    )
    data_sources: Optional[str] = Field(
        description="List of data sources used in string json format."
    )
    user_question: Optional[ValueWithTokens] = Field(
        description="User question."
    )
    # citations: List[str] = Field(
    #     description="List of citations used in the final prompt."
    # )
    context: Optional[ValueWithTokens] = Field(
        description="List of context strings used the final prompt in json format."
    )
    chat_history: Optional[ValueWithTokens] = Field(
        description="Chat history used in the final prompt in json format."
    )
    full_prompt: Optional[ValueWithTokens] = Field(
        description="Final prompt used in the LLM call."
    )
    llm_response: Optional[ValueWithTokens] = Field(
        description="LLM answer."
    )

    class Config:
        schema_extra = {
            "example": {
                "response": "OK",
                "prompt_template": "You are a copilot.\n{context} {history} {answer}",
                # TODO:
                "data_sources": ['{}'],
                "user_question": "Who are you?",
                # "citations": "Who are you?",
                "context": ['[{"page_content": "This is some text from the data files.", "metadata": {"source": "copilots/data/test.txt"}}]'],
                "chat_history": ['[{"user": "Who are you?", {"copilot": "I am a copilot."}]'],
                "full_prompt": "Full formatted prompt text here",
                "llm_response": "I am a copilot.",
            }
        }


class EvaluationInput(BaseModel):
    query: str = Field(
        description="Query sent to the conversation endpoint."
    )
    answer: str = Field(
        description="Answer from the conversation endpoint."
    )
    expected_answer: str = Field(
        description="Expected answer."
    )


class EvaluationResponse(ApiResponse):
    evaluation: str = Field(
        description="Evaluation result, a grade from A-F, with A being the best grade."
    )

    class Config:
        schema_extra = {
            "example": {
                "response": "OK",
                "evaluation": "B",
            }
        }
